"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DocumentContext = void 0;
const assert_1 = __importDefault(require("assert"));
const events_1 = require("events");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
/**
 * @internal
 */
class DocumentContext extends events_1.EventEmitter {
    constructor(routingKey, head, log, getLatestTail) {
        super();
        this.routingKey = routingKey;
        this.log = log;
        this.getLatestTail = getLatestTail;
        this.closed = false;
        this.contextError = undefined;
        this.headUpdatedAfterResume = false; // used to track whether the head has been updated after a resume event, so that we allow moving out of order only once during resume.
        // Head represents the largest offset related to the document that is not checkpointed.
        // Tail will be set to the checkpoint offset of the previous head
        this.headInternal = head;
        this.tailInternal = this.getLatestTail();
    }
    get head() {
        return this.headInternal;
    }
    get tail() {
        return this.tailInternal;
    }
    /**
     * Returns whether or not there is pending work in flight - i.e. the head and tail are not equal
     */
    hasPendingWork() {
        return this.headInternal !== this.tailInternal;
    }
    /**
     * Updates the head offset for the context.
     */
    setHead(head, resumeBackToOffset) {
        (0, assert_1.default)(head.offset > this.head.offset ||
            (head.offset === resumeBackToOffset && !this.headUpdatedAfterResume), `Head offset ${head.offset} must be greater than the current head offset ${this.head.offset} or equal to the resume offset ${resumeBackToOffset} if not yet resumed (headUpdatedAfterResume: ${this.headUpdatedAfterResume}). Topic ${head.topic}, partition ${head.partition}, tenantId ${this.routingKey.tenantId}, documentId ${this.routingKey.documentId}.`);
        // If head is moving backwards
        if (head.offset <= this.head.offset) {
            if (head.offset <= this.tailInternal.offset) {
                server_services_telemetry_1.Lumberjack.info("Not updating documentContext head since new head's offset is <= last checkpoint offset (tailInternal), returning early", {
                    newHeadOffset: head.offset,
                    currentHeadOffset: this.head.offset,
                    tailInternalOffset: this.tailInternal.offset,
                    documentId: this.routingKey.documentId,
                });
                return false;
            }
            server_services_telemetry_1.Lumberjack.info("Allowing the document context head to move to the specified offset", {
                resumeBackToOffset,
                currentHeadOffset: this.head.offset,
                documentId: this.routingKey.documentId,
            });
        }
        // When moving back to a state where head and tail differ we set the tail to be the old head, as in the
        // constructor, to make tail represent the inclusive top end of the checkpoint range.
        if (!this.hasPendingWork()) {
            this.tailInternal = this.getLatestTail();
        }
        if (!this.headUpdatedAfterResume && resumeBackToOffset !== undefined) {
            server_services_telemetry_1.Lumberjack.info("Setting headUpdatedAfterResume to true", {
                resumeBackToOffset,
                currentHeadOffset: this.head.offset,
                documentId: this.routingKey.documentId,
            });
            this.headUpdatedAfterResume = true;
        }
        this.headInternal = head;
        return true;
    }
    checkpoint(message, restartOnCheckpointFailure) {
        if (this.closed) {
            return;
        }
        // Assert offset is between the current tail and head
        const offset = message.offset;
        (0, assert_1.default)(offset > this.tail.offset && offset <= this.head.offset, `Checkpoint offset ${offset} must be greater than the current tail offset ${this.tail.offset} and less than or equal to the head offset ${this.head.offset}. Topic ${message.topic}, partition ${message.partition}, tenantId ${this.routingKey.tenantId}, documentId ${this.routingKey.documentId}.`);
        // Update the tail and broadcast the checkpoint
        this.tailInternal = message;
        this.emit("checkpoint", restartOnCheckpointFailure);
    }
    error(error, errorData) {
        this.contextError = error;
        server_services_telemetry_1.Lumberjack.verbose("Emitting error from documentContext");
        this.emit("error", error, errorData);
    }
    close() {
        this.closed = true;
        this.removeAllListeners();
    }
    getContextError() {
        return this.contextError;
    }
    pause(offset, reason) {
        this.headUpdatedAfterResume = false; // reset this flag when we pause
        this.emit("pause", offset, reason);
    }
    resume() {
        this.emit("resume");
    }
}
exports.DocumentContext = DocumentContext;
//# sourceMappingURL=documentContext.js.map